#include <string.h>
#include "Filters.h"
#include "sampleGrabberCB.h"
#include <wchar.h>

#ifndef _AviHeader
#define _AviHeader
#ifdef MAX_PATH
	#undef MAX_PATH
#endif
#define MAX_PATH 5000

class AviHeader{

	IGraphBuilder *pGraph;
    IMediaControl *pMediaControl;
	IMediaEvent *pMediaEvent;
	IMediaSeeking *pMediaSeeking;
	IMediaFilter *pMediaFilter;
	IPin* pSourceOut[2];
	ISampleGrabber * sampleGrabber;
	GUID time_format;

	char videoFilename[MAX_PATH], outAviVideoFilename[MAX_PATH], errMessage[ERR_BUFFER_LEN];
	WCHAR * wVideoFilename, * wOutAviVideoFilename;
	int grabberPixmapFormat; // enum in sampleGrabberCB.h
	CSampleGrabberCB * sampleGrabberCB;

public:
	LONGLONG num_frames, current_frame;
	bool setupGrabberGraphResult;
	int width, height;
	double frame_rate;
	unsigned char * pixmap;

	void initMembers(){
		pGraph = NULL;
		pMediaControl = NULL;
		pMediaEvent = NULL;
		pMediaSeeking = NULL;
		sampleGrabberCB = NULL;
		wVideoFilename = NULL;
		wOutAviVideoFilename = NULL;
		setupGrabberGraphResult = false;
		frame_rate = 30;  //Set default value
	}

	AviHeader(char * vidFilename, int ingrabberPixmapFormat){

		initMembers();
		grabberPixmapFormat = ingrabberPixmapFormat;

		strcpy_s(videoFilename, MAX_PATH, vidFilename);
		int len = (int)strlen(videoFilename);

		wVideoFilename = new WCHAR[len + 10];
		mbstowcs(wVideoFilename, videoFilename, len + 1);
		setupGrabberGraphResult = setupGrabberGraph();
	}
	
	~AviHeader(){
		if(pGraph != NULL)
		{
			pMediaControl -> Release();
			pMediaEvent -> Release();
			pMediaSeeking -> Release();
			pMediaFilter -> Release();
			pGraph -> Release();
			delete sampleGrabberCB;
			CoUninitialize();
		}
		delete wVideoFilename;
		delete wOutAviVideoFilename;
	}
	
	HRESULT EnumFilters(IBaseFilter *& lastFilter, IBaseFilter *& lastButOneFilter) 
	{
		IEnumFilters *pEnum = NULL;
		IBaseFilter *pFilter = NULL;
		ULONG cFetched;
		lastFilter = lastButOneFilter = NULL;

		HRESULT hr = pGraph->EnumFilters(&pEnum);
		if (FAILED(hr)) return hr;

		//mexPrintf("Filter enumeration \n"); 
		while(pEnum->Next(1, &pFilter, &cFetched) == S_OK)
		{
			FILTER_INFO FilterInfo;
			hr = pFilter->QueryFilterInfo(&FilterInfo);
			if (FAILED(hr))
			{
				DISPLAY_ERROR_BOX("Could not get the filter info", __FILE__, __LINE__)
				continue;  // Maybe the next one will work.
			}

			#ifdef UNICODE
				MessageBox(NULL, FilterInfo.achName, TEXT("Filter Name"), MB_OK);
			#else
				char szName[MAX_FILTER_NAME];
				int cch = WideCharToMultiByte(CP_ACP, 0, FilterInfo.achName,
									MAX_FILTER_NAME, szName, MAX_FILTER_NAME, 0, 0);
				//mexPrintf("%s\n", szName); 
			#endif

			// The FILTER_INFO structure holds a pointer to the Filter Graph
			// Manager, with a reference count that must be released.
			if (FilterInfo.pGraph != NULL)
			{
				FilterInfo.pGraph->Release();
			}

			if( lastFilter == NULL && !strcmp(szName, "Video Renderer") )
				lastFilter = pFilter;
			else if( lastFilter != NULL && lastButOneFilter == NULL )
				lastButOneFilter = pFilter;
			else
				pFilter -> Release();	
		}

		pEnum->Release();
		return S_OK;
	}

	bool setupGrabberGraph(){
		
		CoInitialize(NULL);
		
		if( !getFilterGraphAndStuff((void**) &pGraph, (void**) &pMediaControl, (void**) &pMediaEvent, (void**) &pMediaSeeking, (void**) &pMediaFilter) )
			return 0;

		pMediaFilter->SetSyncSource(NULL);

		if( !addSource(wVideoFilename, pGraph, pSourceOut) ) return 0;

		pGraph -> Render(pSourceOut[0]);
		IBaseFilter * filter1 = NULL, *filter2 = NULL;
		EnumFilters(filter1, filter2);
		pGraph -> RemoveFilter(filter1);
		filter1 -> Release();
		pSourceOut[0] = GetPin(filter2, PINDIR_OUTPUT);
		filter2 -> Release();

		EnumFilters(filter1, filter2);
		sampleGrabber = NULL;
		if( !addVideoSampleGrabber(pGraph, pSourceOut, &sampleGrabber) ) return 0;
		EnumFilters(filter1, filter2);
		
		AM_MEDIA_TYPE mt;
		sampleGrabber -> GetConnectedMediaType( &mt );
		VIDEOINFOHEADER* vih = (VIDEOINFOHEADER*) mt.pbFormat;
		
		sampleGrabberCB = new CSampleGrabberCB( vih->bmiHeader.biWidth, vih->bmiHeader.biHeight, grabberPixmapFormat, pMediaControl);
		sampleGrabber -> SetCallback(sampleGrabberCB, 1);
		sampleGrabber -> SetOneShot( FALSE );
		sampleGrabber -> SetBufferSamples( TRUE );
		
		FreeMediaType( mt );
		
		addFilter(CLSID_NullRenderer, L"Null Renderer", pGraph, pSourceOut, 0);
		
		time_format = TIME_FORMAT_FRAME;
		if( pMediaSeeking -> SetTimeFormat(&time_format) != S_OK )
		{
			time_format = TIME_FORMAT_MEDIA_TIME;
			if( pMediaSeeking -> SetTimeFormat(&time_format) != S_OK )
			{
				DISPLAY_ERROR_BOX("cannot seek given video format using either TIME_FORMAT_FRAME or TIME_FORMAT_MEDIA_TIME", __FILE__, __LINE__)
				return 0;
			}
		}
		
		LONGLONG temp;
		pMediaSeeking -> GetDuration( &temp );
		if( time_format == TIME_FORMAT_MEDIA_TIME )
		{
			int slashindex = (int)strlen(videoFilename) - 1;
			while( slashindex >= 0 && videoFilename[slashindex] != '/' && videoFilename[slashindex] != '\\' )
				slashindex--;
			slashindex++;

			char * videofile = videoFilename + slashindex;

			int minusindex = (int)strlen(videofile) - 1;
			while( minusindex >= 0 && videofile[minusindex] != '-' )
				minusindex--;
			minusindex++;

			char * camerapart = videofile + minusindex;
			int capture_cameraid = 4;
			if( sscanf_s(camerapart, "camera%d.%*s", &capture_cameraid) == 1 && capture_cameraid != 4 )
				frame_rate = 60;

			num_frames = (int)(frame_rate * temp / 1e7  + 1);
			//fprintf(stderr,"%s:%d %d %d %d %lf\n", __FILE__, __LINE__, capture_cameraid, (int)num_frames, (int)temp, frame_rate);
		}
		else
			num_frames = temp;

		current_frame = 0;

		pixmap = sampleGrabberCB -> pixmap;
		width = sampleGrabberCB -> Width;
		height = sampleGrabberCB -> Height;

		return 1;
	}

	void getFrame(LONGLONG frame_number){
		
		if( frame_number > num_frames || frame_number < 0 )
		{
			sprintf_s(errMessage, ERR_BUFFER_LEN, "%s:%d frame_number %d > num_frames %d || frame_number < 0\n", __FILE__, __LINE__, 
						(int)frame_number, (int)num_frames);
			fprintf(stderr, errMessage);
		}
		
		LONGLONG temp;
		if( time_format == TIME_FORMAT_MEDIA_TIME )
		{
			temp = (LONGLONG)(frame_number * 1e7 / frame_rate + 0.5);
		}
		else
			temp = frame_number;
		
		HRESULT hr;
		if( (hr = pMediaSeeking -> SetPositions(&temp, AM_SEEKING_AbsolutePositioning, NULL, AM_SEEKING_NoPositioning)) != S_OK )
			DISPLAY_ERROR_BOX(hr, __FILE__,__LINE__)

		pMediaControl -> Run();

		long evCode;
		pMediaEvent -> WaitForCompletion(INFINITE, &evCode);
	}

	void writePixmapToPPM(char * filename)
	{
		int h = height;
		int w = width;
		/* Write out ppm for debugging */
		char errMessage[1000];

		if( grabberPixmapFormat != PPM_FORMAT_PIXMAP && grabberPixmapFormat != BGR_FORMAT_PIXMAP)
		{
			sprintf_s(errMessage, 1000, "%s:%d writePixmapToPPM only for PPM_FORMAT_PIXMAP \n", __FILE__, __LINE__, filename);
			fprintf(stderr, errMessage);
			return;
		}
		FILE * fp = fopen(filename,"wb");
		if( fp == NULL )
		{
			sprintf_s(errMessage, 1000, "%s:%d unable to open file %s for writing \n", __FILE__, __LINE__, filename);
			fprintf(stderr, errMessage);
			return;
		}
		fprintf(fp,"P6\n%d %d\n255\n",w,h);
		fwrite(pixmap, w*h*3, 1, fp);
		fclose(fp);
	}

};

#endif
