#ifndef AVISOCKET_H
#define AVISOCKET_H

#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#undef MS_WIN
#ifdef WIN32 
	#define MS_WIN
#endif
#ifdef WIN64 
	#define MS_WIN
#endif

#ifndef MS_WIN 
	#include <unistd.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <netdb.h>
	#define CLOSESOCKET close
#else
	#define CLOSESOCKET closesocket
#endif

#ifdef mex_h
#	define my_error(str) mexPrintf(str)
#	define my_printf(str) mexWarnMsgIdAndTxt("netAvi:info", str)
#else
#	define my_error(str) fprintf(stderr, str)
#	define my_printf(str) fprintf(stdout, str)
#endif

static char output_str[1000];

static unsigned int buflen = 10000000;
#define CONNECTION_TESTLEN 99999999

#define my_min(a,b) ( ((a) < (b))? (a):(b) )

class AviSocket{
	public:
	int port_number, server_or_client;
	char * buffer, ip_address[100];
	int sockfd,  accept_sockfd;
	unsigned int data_len;
	struct sockaddr_in address;
	AviSocket(){
		init();
	}
	void init(){
		buffer = new char[buflen];
		sockfd = accept_sockfd = -1;
	}
	AviSocket(AviSocket * server_socket){
		init();
		sockfd = server_socket -> sockfd;
		accept_sockfd = server_socket -> accept_sockfd;
		server_or_client = server_socket -> server_or_client;
	}
	AviSocket(int input_port_number, char * input_ip_address){
		init();
		this -> port_number = input_port_number;
		if ((sockfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
			sprintf(output_str, "%s:%d Error creating unnamed socket %s:%d\n%s\n",
					__FILE__, __LINE__, input_ip_address, port_number, strerror(errno));
			my_error(output_str);
			exit (1);
		}
		if( strlen(input_ip_address) == 0 ){
			server_or_client = 1;
			address.sin_addr.s_addr = INADDR_ANY;
		}else{
			server_or_client = 0;
			strcpy(ip_address, input_ip_address);
			//struct hostent * hostEntry = gethostbyaddr(ip_address, strlen(ip_address), AF_INET);
			struct hostent * hostEntry = gethostbyname(ip_address);
			if( !hostEntry ){
				sprintf(output_str,"%s:%d Client: Error getting host entry for %s\n%s\n",
					__FILE__, __LINE__, ip_address, strerror(errno));
				my_error(output_str);
				return;
			}
			memcpy( &(address.sin_addr.s_addr), hostEntry -> h_addr, hostEntry->h_length);
			//address.sin_addr.s_addr = inet_addr(ip_address);
		}
		address.sin_family = AF_INET;
		address.sin_port = htons(port_number);
		if( server_or_client ){
			if (bind(sockfd, (struct sockaddr *)&address, sizeof(address)) < 0) {
				sprintf(output_str,"%s:%d Server: Error binding to :%d\n%s\n", __FILE__, __LINE__, port_number, strerror(errno));
				my_error(output_str);
				exit (1);
			} else {
				sprintf(output_str, "%s:%d Server socket openend at :%d\n", __FILE__, __LINE__, port_number);
				my_printf(output_str);
			}
		}else{
			if (connect (sockfd, (struct sockaddr *)&address, sizeof(address)) < 0) {
				sockfd = -1;
				sprintf(output_str,"%s:%d Client: Error connecting to %s:%d\n%s\n",
					__FILE__, __LINE__, ip_address, port_number, strerror(errno));
				my_error(output_str);
			}
		}
	}
	void close_socket(){
		if( server_or_client ){
			CLOSESOCKET(accept_sockfd);
			accept_sockfd = -1;
		}
		else{
			CLOSESOCKET(sockfd);
			sockfd = -1;
		}
	}
	void cleanUp(){
		close_socket();
		delete [] buffer;
		buffer = NULL;
	}
	~AviSocket(){
		cleanUp();
	}
	int server_wait_for_client(){
		if( listen(sockfd, 3) < 0){
			sprintf(output_str,"%s:%d Listen error %s\n", __FILE__, __LINE__, strerror(errno));
			my_error(output_str);
			return 0;
		}
		accept_sockfd = accept(sockfd, (sockaddr *)NULL, NULL);
		if( accept_sockfd < 0){
			sprintf(output_str,"%s:%d Accept error %s\n", __FILE__, __LINE__, strerror(errno));
			my_error(output_str);
			return 0;
		}
		return 1;	
	}
	int write_test_connection(){
		int use_sockfd;
		use_sockfd = (server_or_client)? accept_sockfd : sockfd;
		unsigned int test_val = htonl(CONNECTION_TESTLEN);
		int tlen = send(use_sockfd, (char*)&test_val, sizeof(int), 0);
		if( tlen == sizeof(int) )
			return 1;
		
		sprintf(output_str, "%s:%d Send error %s %d of %d bytes \n%s\n", __FILE__, __LINE__,
				strerror(errno), tlen, (int)sizeof(int), strerror(errno));
		my_error(output_str);
		return 0;
	}
	int write_to_socket(){
		int use_sockfd;
		use_sockfd = (server_or_client)? accept_sockfd : sockfd;
		if( data_len > buflen ){
			sprintf(output_str,"%s:%d Send error data_len %u > buflen %u \n%s\n", __FILE__, __LINE__,
					data_len, buflen, strerror(errno));
			my_error(output_str);
			return 0;
		}
		unsigned int net_len = htonl(data_len);
		int tlen = send(use_sockfd, (char*)&net_len, sizeof(int), 0);
		char * tbuf = buffer;
		unsigned int sent_len = 0;
		while( tlen >=0 && sent_len < data_len ){
			tlen = send(use_sockfd, tbuf, my_min(data_len - sent_len,1024), 0);
			if( tlen < 0 )
				break;
			sent_len += tlen;	
			tbuf += tlen;
		}
		if( tlen < 0 || sent_len != data_len){
			sprintf(output_str,"%s:%d Send error %s\n %u of %u bytes \n%s\n", __FILE__, __LINE__,
				strerror(errno), sent_len, data_len, strerror(errno));
			my_error(output_str);
			return 0;
		}
		//sprintf(output_str,"%s:%d Sent %d bytes\n",__FILE__, __LINE__, data_len); my_printf(output_str);
		return 1;
	}
	int read_from_socket(){
		int use_sockfd, tlen;
		char * tbuf = buffer;
		use_sockfd = (server_or_client)? accept_sockfd : sockfd;
		unsigned int net_len;
		while(1){
			tlen = recv(use_sockfd, (char*)&net_len, sizeof(int), 0);
			if(tlen <= 0){
				sprintf(output_str,"%s:%d Recieve error %s\n", __FILE__, __LINE__, strerror(errno));
				my_error(output_str);
				return 0;
			}
			data_len = ntohl(net_len);
			if( data_len != CONNECTION_TESTLEN )
				break;
		}
		if( data_len > buflen ){
			sprintf(output_str,"%s:%d Recieve error data_len %u > buflen %u \n%s\n", __FILE__, __LINE__,
					data_len, buflen, strerror(errno));
			my_error(output_str);
			return 0;
		}
		if( data_len == 0 ){
			sprintf(output_str,"%s:%d Server error data_len  = 0\n%s\n", __FILE__, __LINE__, strerror(errno));
			my_error(output_str);
			return 0;
		}
		for(unsigned int i = 0; i< buflen; i++) buffer[i] = 0;
		//bzero(buffer, buflen);
		unsigned int read_len  = 0;
		while( read_len < data_len ){
			tlen = recv(use_sockfd, tbuf, 1024, 0);
			if( tlen <= 0 ) break;
			tbuf += tlen;
			read_len += tlen;
			//my_printf("%d of %d\n", read_len, data_len);
		}
		if(tlen < 0 || read_len < data_len ){
			sprintf(output_str,"%s:%d Recieve error %s\n", __FILE__, __LINE__, strerror(errno));
			my_error(output_str);
			sprintf(output_str,"%s:%d Recieved only %u of %u bytes \n", __FILE__, __LINE__,
				read_len, data_len);
			my_error(output_str);
			return 0;
		}
		//sprintf(output_str,"%s:%d Recieved %d bytes\n",__FILE__, __LINE__, data_len); my_printf(output_str);
		return 1;
	}
	void write_empty_to_socket(){
		data_len = 0;
		write_to_socket();
	}
	int write_filename_to_socket(char *& filename){
		char * tbuf = buffer;
		int tsize = 0;
		tsize = strlen(filename)+1; memcpy(tbuf, filename, tsize); tbuf += tsize;
		data_len = tbuf - buffer; 
		return write_to_socket();
	}
	int read_filename_from_socket(char * filename){
		if(!read_from_socket())
			return 0;
		char * tbuf = buffer;
		int tsize = 0;
		tsize = data_len; memcpy(filename, tbuf, tsize); tbuf += tsize;
		return 1;
	}
	int write_framenumber_to_socket(int & frame_number){
		char * tbuf = buffer;
		int tsize = 0;
		tsize = sizeof(int); memcpy(tbuf, &frame_number, tsize); tbuf += tsize;
		data_len = tbuf - buffer; 
		return write_to_socket();
	}
	int read_framenumber_from_socket(int & frame_number){
		if(!read_from_socket())
			return 0;
		char * tbuf = buffer;
		int tsize = 0;
		tsize = sizeof(int); memcpy(&frame_number, tbuf, tsize); tbuf += tsize;
		return 1;
	}
	int write_aviinfo_to_socket( int &w, int &h, int &nframes, double &fps, double &total_time_sec){
		char * tbuf = buffer;
		int tsize = 0;
		tsize = sizeof(int); memcpy(tbuf, &w, tsize); tbuf += tsize;
		tsize = sizeof(int); memcpy(tbuf, &h, tsize); tbuf += tsize;
		tsize = sizeof(int); memcpy(tbuf, &nframes, tsize); tbuf += tsize;
		tsize = sizeof(double); memcpy(tbuf, &fps, tsize); tbuf += tsize;
		tsize = sizeof(double); memcpy(tbuf, &total_time_sec, tsize); tbuf += tsize;
		data_len = tbuf - buffer; 
		return write_to_socket();
	}
	int read_aviinfo_from_socket( int &w, int &h, int &nframes, double &fps, double &total_time_sec){
		if(!read_from_socket())
			return 0;
		char * tbuf = buffer;
		int tsize = 0;
		tsize = sizeof(int); memcpy(&w, tbuf, tsize); tbuf += tsize;
		tsize = sizeof(int); memcpy(&h, tbuf, tsize); tbuf += tsize;
		tsize = sizeof(int); memcpy(&nframes, tbuf, tsize); tbuf += tsize;
		tsize = sizeof(double); memcpy(&fps, tbuf, tsize); tbuf += tsize;
		tsize = sizeof(double); memcpy(&total_time_sec, tbuf, tsize); tbuf += tsize;
		return 1;
	}
	int write_frame_to_socket( int &w, int &h, unsigned char *& pixmap){
		char * tbuf = buffer;
		int tsize = 0;
		tsize = sizeof(int); memcpy(tbuf, &w, tsize); tbuf += tsize;
		tsize = sizeof(int); memcpy(tbuf, &h, tsize); tbuf += tsize;
		tsize = w*h*3; memcpy(tbuf, pixmap, tsize); tbuf += tsize;
		data_len = tbuf - buffer; 
		return write_to_socket();
	}
	int read_frame_from_socket( int &w, int &h, unsigned char *& pixmap){
		if(!read_from_socket())
			return 0;
		char * tbuf = buffer;
		int tsize = 0;
		tsize = sizeof(int); memcpy(&w, tbuf, tsize); tbuf += tsize;
		tsize = sizeof(int); memcpy(&h, tbuf, tsize); tbuf += tsize;
		pixmap = (unsigned char *)tbuf;
		return 1;
	}

};

#endif
