#include <mex.h>
#ifdef WIN32
#include <windows.h>
#endif
#ifdef WIN64
#include <windows.h>
#endif
#include "AviSocket.h"
#include <math.h>

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[]){
					  
	char  *buffer;
	int buflen, frame_no;

	if( !mxIsChar(prhs[0]) )
		mexErrMsgTxt("input filename is not a string\n");
	buflen = mxGetN(prhs[0])*sizeof(mxChar)+1;
	char * file_name = new char[buflen+10];
	mxGetString(prhs[0], file_name, buflen);

	char server_ip_address[10000];
	if( !mxIsChar(prhs[1]) )
		mexErrMsgTxt("input server ip address is not a string\n");
	mxGetString(prhs[1], server_ip_address, mxGetN(prhs[1])*sizeof(mxChar)+1);

	if( !mxIsDouble(prhs[2]) )
		mexErrMsgTxt("input server port_number is not a number\n");
	int	port_number = (int)floor(mxGetScalar(prhs[2]) + 0.1);

	AviSocket * aviSocket = new AviSocket(port_number, server_ip_address);
	if( aviSocket -> sockfd < 0 )
		mexErrMsgTxt("Error connecting to server\n");

	aviSocket -> write_filename_to_socket(file_name); 

	buflen = sizeof(AviSocket);
	plhs[0] = mxCreateDoubleMatrix(buflen/sizeof(double)+1, 1, mxREAL);
	buffer = (char*)mxGetPr(plhs[0]);
	memcpy(buffer, aviSocket, buflen);
	
	int w, h, nframes;
	double fps, total_time_sec;
	if( aviSocket -> read_aviinfo_from_socket(w, h, nframes, fps, total_time_sec) == 0){
		char err_msg[1000];
		sprintf(err_msg, "Server failed to open %s.\n Check server log\n", file_name);
		mexErrMsgTxt(err_msg);
	}

	plhs[1] = mxCreateDoubleMatrix(5, 1, mxREAL);
	double * out = mxGetPr(plhs[1]);
	out[0] = w;
	out[1] = h;
	out[2] = nframes;
	out[3] = fps;
	out[4] = total_time_sec;
	
	delete file_name;
}

